#ifndef clean_CLEANUP_HPP_INCLUDED
#define clean_CLEANUP_HPP_INCLUDED 1

#include <algorithm> // for_each()
#include <stdexcept> // runtime_error
#include <list>
#include <map>

/**
   The 'clean' namespace encapsulates a small utility library for
   deallocating unmanaged pointers. In particular, it is designed to
   take over the task of deallocating pointers contained in standard
   containers, but the general approach can apply to other cases as
   well.

   Motivation:

   Have you ever wanted/needed to clean up the contents of a nested
   container, like:

   list<map<int,T*> >

   But you don't want to walk the container(s) or invoke for_each()
   and various functors? In some cases, you can't even walk the
   containers because your algorithm doesn't know they are containers!
   (That also makes them unsafe to simply delete.)

   That's what 'clean' is about: automating that. How about simply:

   clean::cleanup( mylist );


   Author: stephan@s11n.net

   License: Public Domain


   July 2005: initial code written for libs11n 1.1.3.

*/
namespace clean {

	/** Default cleanup functor impl for cleanup_traits. */
	template <typename T>
	struct default_cleanup_functor
	{
		typedef T cleaned_type;
		/** Does nothing. */
		void operator()( cleaned_type & ) throw()
		{
		}
	};

	/**
	   This is the core type of the clean framework.  It's
	   cleanup_functor is responsible for defining rules for
	   "cleaning up" any objects contained in an object of type T.
	   
	   "Cleaning up" means, basically, apply recursive cleanup
	   functors to any child objects. On "leaf" objects - those
	   which manage their own pointers - we do nothing at all.

	*/
	template <typename T>
	struct cleanup_traits
	{
		typedef T cleaned_type;
		typedef default_cleanup_functor<T> cleanup_functor;
	};


	/**
	   General specialization to treat (T*) as (T) for cleanup_traits
	   typing purposes.
	*/
	template <typename T>
	struct cleanup_traits<T*> : public cleanup_traits<T>
	{
	};

	/**
	   This is the main entrypoint into the cleanup service.

	   Uses cleanup_traits<T>::cleanup_functor to "clean up" t.

	   Requirements for T:

	   - Must either be cleanable with the default cleanup functor
	   (which means it has no special cleaning requirements) or
	   must have a working specialization of either
	   cleanup_traits<T> or default_cleanup_functor<T> installed.
	*/
	template <typename T>
	void cleanup( T & t ) throw()
	{
		try
		{
			typedef typename cleanup_traits<T>::cleanup_functor CF;
			CF cf;
			cf( t );
		}
		catch(...)
		{
		}
	}

	/**
	   Overloaded to accept pointers. Does the equivalent of:

	   1) cleanup<T>( *t )

	   2) delete t

	   3) t = 0


	   General requirements for T:

	   - As for the non-pointer variant.

	   - Must be safe to delete it.

	   - T's dtor must not throw.

	   After calling this, (t==0).
	*/
 	template <typename T>
 	void cleanup( T * & t ) throw()
 	{
 		if( t )
 		{
  			cleanup<T>( *t );
  			delete t;
  			t = 0;
 		}
 	}


	/**
	   An auto_ptr-like type which is intended to simplify
	   pointer/exception safety in some algorithms.

	   Requirements on T are the same as for cleanup<T>(T*&).
	*/
	template <typename T>
	struct cleanup_ptr
	{
	public:
		/**
		   The same as T.
		*/
		typedef T cleaned_type;
	private:
		cleaned_type * m_ptr;
		cleanup_ptr & operator=( const cleanup_ptr & ); // Not Implemented
		cleanup_ptr( const cleanup_ptr & ); // Not Implemented
		void do_cleanup() throw()
		{
			if( this->m_ptr )
			{
				cleanup<cleaned_type>( this->m_ptr );
			}
		}
	public:
		/**
		   Constructs an object pointing to nothing.
		*/
		cleanup_ptr() : m_ptr(0)
		{
		}
		/**
		   Transfers ownership of p to this object.
		 */
		cleanup_ptr( cleaned_type * p ) : m_ptr(p)
		{
		}

		/**
		   Uses clean_traits<cleaned_type>::cleanup_functor to
		   free up up this object's pointed-to object.
		*/
		~cleanup_ptr() throw()
		{
			this->do_cleanup();
		}
		/**
		   Dereferences this object's pointed-to object.
		   If this object does not point to anything it
		   throws a std::runtime_error.
		 */
		cleaned_type & operator*() throw(std::runtime_error)
		{
			if( ! this->m_ptr )
			{
				throw std::runtime_error("Attempt to dereference a null pointer via cleanup_ptr<>::operator*()" );
			}
			return *this->m_ptr;
		}

		/**
		   Returns this object's pointed-to object without
		    transfering ownership.
		*/
		cleaned_type * get()
		{
			return this->m_ptr;
		}

		/**
		   Passes ownership of p to this object. This takes
		   the place of a copy/assign operators, since those
		   conventionally take a const right-hand argument.

		   Destroys the object this object pointed to before
		   taking over ownership.
		*/
		void take( cleaned_type * p )
		{
			if( p != this->m_ptr )
			{
				this->do_cleanup();
				this->m_ptr = p;
			}
		}

		/**
		   Transfers ownership of the pointed-to object to the
		   caller.
		*/
		cleaned_type * release()
		{
			cleaned_type * x = this->m_ptr;
			this->m_ptr = 0;
			return x;
		}

		/**
		   Swaps ownership of pointers with rhs.
		*/
		void swap( cleanup_ptr & rhs )
		{
			cleaned_type * x = this->m_ptr;
			this->m_ptr = rhs.m_ptr;
			rhs.m_ptr = x;
		}
	};
	


	/** Intended for use with for_each(). */
	struct do_cleanup
	{
		/**
		   Calls cleanup<T>(t)
		*/
		template <typename T>
		void operator()( T & t ) throw()
		{
			cleanup( t );
		}
		/**
		   Calls cleanup<T>(t).
		*/
		template <typename T>
		void operator()( T * & t ) throw()
		{
			// CERR << "do_cleanup(T*&)\n";
			cleanup<T>( t );
		}
	};


	/**
	   Sample specialization of a cleanup functor for std::maps.
	   The same code will work for std::multimap and many map-style
	   types.
	*/
	template <typename KT,typename MT>
	struct default_cleanup_functor< std::map< KT, MT > >
	{
		typedef std::map< KT, MT > cleaned_type;
		typedef KT key_type;
		typedef MT mapped_type;
		/**
		   For each item in m, it's value is "cleaned
		   up". Achtung: keys are const, so we cannot clean
		   them up!

		   After calling this, m is empty.
		*/
		void operator()( cleaned_type & m )
		{
			typedef typename cleaned_type::iterator IT;
			IT b = m.begin();
			IT e = m.end();
			if( e == b ) return;
			typedef typename cleanup_traits<MT>::cleaned_type MTBase; // stripping pointer part
			for( ; e != b; ++b )
			{
				cleanup<MTBase>( (*b).second ); // this is why we wanted to strip any pointer part
			}
			m.clear();
		}
	};

	/**
	   Sample specialization of a cleanup functor for std::list.
	   The same code will work on most list-style types.

	*/
	template <typename VT>
	struct default_cleanup_functor< std::list< VT > >
	{
		typedef std::list< VT > cleaned_type;
		/**
		   "Cleans up" each element in c.

		   After calling this, c is empty.
		 */
		void operator()( cleaned_type & c ) throw()
		{
			std::for_each( c.begin(), c.end(), do_cleanup() );
			c.clear();
		}
	};

} // namespace clean


#endif // clean_CLEANUP_HPP_INCLUDED
